#!/usr/bin/env python3

"""Corn jobs 🌽

Supporting the core app with asynchronous maintenance tasks.

Mostly related to monitoring and managing Printer state.
"""

from contextlib import closing
from datetime import datetime, timedelta
import logging
from pathlib import Path
from threading import Event, Thread
from time import sleep
from typing import Callable
from urllib import parse as urlparse

from fastmail import FastMailSMTP
from flask import Flask as App, render_template
from octorest import OctoRest as _OR
from requests import Response
from requests.exceptions import (
    ConnectionError,
    HTTPError,
    Timeout,
)
from tentacles.store import Store


class OctoRest(_OR):
    def _get(self, path, params=None):
        url = urlparse.urljoin(self.url, path)
        response = self.session.get(url, params=params, timeout=(1.0, 1.0))
        self._check_response(response)

        return response.json()

    def _check_response(self, response: Response):
        response.raise_for_status()
        return response


log = logging.getLogger(__name__)

SHUTDOWN = Event()


def corn_job(every: timedelta):
    def _decorator(f):
        def _helper(*args, **kwargs):
            last = None
            while not SHUTDOWN.is_set():
                if not last or (datetime.now() - last) > every:
                    log.debug(f"Ticking job {f.__name__}")
                    try:
                        last = datetime.now()
                        f(*args, **kwargs)
                    except Exception:
                        log.exception(f"Error while procesing task {f.__name__}")
                else:
                    sleep(1)

        return _helper

    return _decorator


def poll_printers(app: App, store: Store) -> None:
    """Poll printers for their status."""

    for printer in store.list_printers():
        mapped_job = store.fetch_job_by_printer(printer.id)

        def _set_status(status: str):
            if printer.status != status:
                print(f"Printer {printer.id} {printer.status} -> {status}")
                store.update_printer_status(printer.id, status)

        try:
            client = OctoRest(url=printer.url, apikey=printer.api_key)
            printer_job = client.job_info()
            try:
                printer_state = client.printer().get("state").get("flags", {})
            except HTTPError:
                printer_state = {"disconnected": True}

            if printer_state.get("error"):
                # If there's a mapped job, we manually fail it so that
                # there's no possibility of a sync problem between the
                # polling tasks. This violates separation of concerns a bit,
                # but appears required for correctness.
                if mapped_job:
                    store.finish_job(mapped_job.id, "error")

                _set_status("error")

            elif printer_state.get("disconnected"):
                _set_status("disconnected")

            elif printer_state.get("printing"):
                _set_status("running")

            elif printer_job.get("state").lower() == "connecting":
                _set_status("connecting")

            elif printer_state.get("ready"):
                _set_status("idle")

            else:
                raise Exception(
                    f"Indeterminate state {printer_job!r} {printer_state!r}"
                )

        except (ConnectionError, Timeout):
            _set_status("error")

        except HTTPError as e:
            assert isinstance(e.response, Response)
            if e.response.status_code in [403, 401] or "error" in printer_job:
                _set_status("error")

            elif e.response.json().get("error") == "Printer is not operational":
                _set_status("disconnected")

            else:
                raise Exception(
                    f"Indeterminate state {e.response.status_code}, {e.response.json()!r}"
                )


def assign_jobs(app: App, store: Store) -> None:
    """Assign jobs to printers. Uploading files and job state management is handled separately."""

    for printer_id in store.list_idle_printers():
        if job_id := store.poll_job_queue():
            store.assign_job(job_id, printer_id)
            print(f"Mapped job {job_id} to printer {printer_id}")


def push_jobs(app: App, store: Store) -> None:
    """Ensure that job files are uploaded and started to the assigned printer."""

    for job in store.list_mapped_jobs():
        printer = store.fetch_printer(job.printer_id)
        file = store.fetch_file(job.user_id, job.file_id)
        if not file:
            log.error(f"Job {job.id} no longer maps to a file")
            store.delete_job(job.user_id, job.id)

        try:
            client = OctoRest(url=printer.url, apikey=printer.api_key)
            printer_job = client.job_info()
            try:
                printer_state = client.printer().get("state").get("flags", {})
            except HTTPError:
                printer_state = {"error": printer_job.get("error")}

            if printer_state.get("error"):
                print(f"Printer {printer.id} is in error, can't push")
                continue

            try:
                client.upload(file.path)
            except HTTPError as e:
                if e.response.status_code == 409:
                    pass
                else:
                    raise

            client.select(Path(file.path).name)
            client.start()
            store.start_job(job.id)
        except TimeoutError:
            pass

        except Exception:
            log.exception("Oop")


def revoke_jobs(app: App, store: Store) -> None:
    """Ensure that job files are uploaded and started to the assigned printer.

    Note that this will ALSO cancel jobs out of the print queue.

    """

    for job in store.list_cancelled_jobs():
        if job.printer_id:
            printer = store.fetch_printer(job.printer_id)

            try:
                print(f"Cancelling running job {job.id}")
                client = OctoRest(url=printer.url, apikey=printer.api_key)
                try:
                    client.cancel()
                except HTTPError as e:
                    if e.response.status_code == 409:
                        pass
                    else:
                        raise

                print(f"Job {job.id} -> cancelled")
                store.finish_job(job.id, "cancelled")

            except TimeoutError:
                pass

            except Exception:
                log.exception("Oop")

        else:
            print(f"Unmapped job {job.id} became cancelled")
            store.finish_job(job.id, "cancelled")


def pull_jobs(app: App, store: Store) -> None:
    """Poll the state of mapped printers to control jobs."""

    for job in store.list_running_jobs():
        printer = store.fetch_printer(job.printer_id)
        try:
            client = OctoRest(url=printer.url, apikey=printer.api_key)
            job_state = client.job_info()
            try:
                printer_state = client.printer().get("state").get("flags", {})
            except HTTPError:
                printer_state = {"disconnected": True, "error": True}

            if printer_state.get("printing"):
                pass

            elif job_state.get("progress", {}).get("completion", 0.0) == 100.0:
                print(f"Job {job.id} has succeeded")
                store.finish_job(job.id, "success")

            elif printer_state.get("error"):
                print(f"Job {job.id} has failed")
                store.finish_job(job.id, "failed")

            elif printer_state.get("cancelling"):
                print(f"Job {job.id} has been acknowledged as cancelled")
                store.finish_job(job.id, "cancelled")

            else:
                print(
                    f"Job {job.id} is in a weird state {job_state.get('progress')!r} {printer_state!r}"
                )

        except TimeoutError:
            pass

        except Exception:
            log.exception("Oop")


def send_emails(app, store: Store):
    with closing(
        FastMailSMTP(
            app.config.get("fastmail", {}).get("username"),
            app.config.get("fastmail", {}).get("key"),
        )
    ) as fm:
        for message in store.poll_spool():
            fm.send_message(
                from_addr="root@tirefireind.us",
                to_addrs=[message.to],
                subject=message.subject,
                msg=message.body,
            )
            store.send_email(message.id)


@corn_job(timedelta(seconds=5))
def run_worker(app: App, db_factory):
    with app.app_context(), closing(db_factory(app)) as store:
        poll_printers(app, store)
        assign_jobs(app, store)
        push_jobs(app, store)
        revoke_jobs(app, store)
        pull_jobs(app, store)
        send_emails(app, store)


def create_workers(app, db_factory: Callable[[App], Store]) -> Event:
    Thread(target=run_worker, args=[app, db_factory]).start()

    return SHUTDOWN
